@@ -9,6 +9,7 @@
 #include "optional.hh"
 #include "option_types.hh"
 #include "ranges.hh"
+#include "regex.hh"
 #include "register_manager.hh"
 #include "shell_manager.hh"
 #include "utils.hh"
@@ -91,11 +92,25 @@ Reader& Reader::operator++()
 {
     kak_assert(pos < str.end());
     if (*pos == '\n')
+    {
         ++line;
+        line_start = ++pos;
+        return *this;
+    }
     utf8::to_next(pos, str.end());
     return *this;
 }
 
+void Reader::next_byte()
+{
+    kak_assert(pos < str.end());
+    if (*pos++ == '\n')
+    {
+        ++line;
+        line_start = pos;
+    }
+}
+
 namespace
 {
 
@@ -151,10 +166,10 @@ QuotedResult parse_quoted_balanced(Reader& reader, char opening_delimiter,
         else if (c == closing_delimiter and level-- == 0)
         {
             auto content = reader.substr_from(start);
-            ++reader.pos;
+            reader.next_byte();
             return {String{String::NoCopy{}, content}, true};
         }
-        ++reader.pos;
+        reader.next_byte();
     }
     return {String{String::NoCopy{}, reader.substr_from(start)}, false};
 }
@@ -178,7 +193,7 @@ String parse_unquoted(Reader& reader)
             else
                 return str;
         }
-        ++reader.pos;
+        reader.next_byte();
     }
     if (beg < reader.str.end())
         str += reader.substr_from(beg);
@@ -284,12 +299,12 @@ Token parse_percent_token(Reader& reader, bool throw_on_unterminated)
     }
 }
 
-auto expand_option(Option& opt, std::true_type)
+auto expand_option(const Option& opt, std::true_type)
 {
     return opt.get_as_string(Quoting::Raw);
 }
 
-auto expand_option(Option& opt, std::false_type)
+auto expand_option(const Option& opt, std::false_type)
 {
     return opt.get_as_strings();
 }
@@ -462,10 +477,9 @@ String expand(StringView str, const Context& context,
 
 String expand(StringView str, const Context& context,
               const ShellContext& shell_context,
-              const std::function<String (String)>& postprocess)
+              const FunctionRef<String (String)>& postprocess)
 {
-    return expand_impl(str, context, shell_context,
-                       [&](String s) { return postprocess(std::move(s)); });
+    return expand_impl(str, context, shell_context, postprocess);
 }
 
 struct command_not_found : runtime_error
@@ -676,6 +690,11 @@ Completions CommandManager::complete(const Context& context,
 
     switch (token.type)
     {
+    case Token::Type::RegisterExpand:
+        return {start , cursor_pos,
+                RegisterManager::instance().complete_register_name(
+                    token.content, cursor_pos_in_token) };
+
     case Token::Type::OptionExpand:
         return {start , cursor_pos,
                 GlobalScope::instance().option_registry().complete_option_name(
@@ -690,6 +709,13 @@ Completions CommandManager::complete(const Context& context,
                 ShellManager::instance().complete_env_var(
                     token.content, cursor_pos_in_token) };
 
+    case Token::Type::FileExpand:
+    {
+        const auto& ignored_files = context.options()["ignored_files"].get<Regex>();
+        return {start , cursor_pos, complete_filename(
+                token.content, ignored_files, cursor_pos_in_token, FilenameFlags::Expand) };
+    }
+
     case Token::Type::Raw:
     case Token::Type::RawQuoted:
     case Token::Type::RawEval:
