@@ -3,6 +3,7 @@
 #include "alias_registry.hh"
 #include "client.hh"
 #include "face_registry.hh"
+#include "buffer_manager.hh"
 #include "register_manager.hh"
 #include "window.hh"
 
@@ -171,7 +172,12 @@ void Context::change_buffer(Buffer& buffer)
     if (has_buffer() and m_edition_level > 0)
        this->buffer().commit_undo_group();
 
-    m_window.reset();
+    if (has_buffer())
+    {
+        auto* current = &this->buffer();
+        m_last_buffer = contains(BufferManager::instance(), current) ? current : nullptr;
+    }
+
     if (has_client())
     {
         client().info_hide();
@@ -179,12 +185,30 @@ void Context::change_buffer(Buffer& buffer)
         client().change_buffer(buffer);
     }
     else
+    {
+        m_window.reset();
         m_selections = SelectionList{buffer, Selection{}};
+    }
 
     if (has_input_handler())
         input_handler().reset_normal_mode();
 }
 
+void Context::forget_buffer(Buffer& buffer)
+{
+    m_jump_list.forget_buffer(buffer);
+    if (m_last_buffer.get() == &buffer)
+        m_last_buffer = nullptr;
+
+    if (&this->buffer() != &buffer)
+        return;
+
+    if (is_editing() && has_input_handler())
+        input_handler().reset_normal_mode();
+
+    change_buffer(m_last_buffer ? *m_last_buffer : BufferManager::instance().get_first_buffer());
+}
+
 SelectionList& Context::selections()
 {
     if (not m_selections)
