@@ -92,8 +92,8 @@ struct MouseHandler
         Buffer& buffer = context.buffer();
         BufferCoord cursor;
         auto& selections = context.selections();
-        const auto key_modifier = (Key::Modifiers)(key.modifiers & Key::Modifiers::MouseEvent);
-        switch (key_modifier)
+        constexpr auto modifiers = Key::Modifiers::Control | Key::Modifiers::Alt | Key::Modifiers::Shift;
+        switch ((key.modifiers & ~modifiers).value)
         {
         case Key::Modifiers::MousePressRight:
             m_dragging = false;
@@ -137,14 +137,8 @@ struct MouseHandler
             selections.sort_and_merge_overlapping();
             return true;
 
-        case Key::Modifiers::MouseWheelDown:
-            m_dragging = false;
-            scroll_window(context, 3);
-            return true;
-
-        case Key::Modifiers::MouseWheelUp:
-            m_dragging = false;
-            scroll_window(context, -3);
+        case Key::Modifiers::Scroll:
+            scroll_window(context, static_cast<int32_t>(key.key), not m_dragging);
             return true;
 
         default: return false;
@@ -209,8 +203,6 @@ public:
             context().hooks_disabled().unset();
             m_hooks_disabled = false;
         }
-
-        context().hooks().run_hook(Hook::NormalBegin, "", context());
     }
 
     void on_disabled(bool temporary) override
@@ -223,8 +215,6 @@ public:
             context().hooks_disabled().unset();
             m_hooks_disabled = false;
         }
-
-        context().hooks().run_hook(Hook::NormalEnd, "", context());
     }
 
     void on_key(Key key) override
@@ -232,16 +222,6 @@ public:
         kak_assert(m_state != State::PopOnEnabled);
         ScopedSetBool set_in_on_key{m_in_on_key};
 
-        // Hack to parse keys sent by terminals using the 8th bit to mark the
-        // meta key. In normal mode, give priority to a potential alt-key than
-        // the accentuated character.
-        if (not (key.modifiers & Key::Modifiers::MouseEvent) and
-            key.key >= 127 and key.key < 256)
-        {
-            key.modifiers |= Key::Modifiers::Alt;
-            key.key &= 0x7f;
-        }
-
         bool do_restore_hooks = false;
         auto restore_hooks = on_scope_end([&, this]{
             if (m_hooks_disabled and enabled() and do_restore_hooks)
@@ -253,8 +233,6 @@ public:
 
         const bool transient = context().flags() & Context::Flags::Draft;
 
-        auto cp = key.codepoint();
-
         if (m_mouse_handler.handle_key(key, context()))
         {
             context().print_status({});
@@ -264,7 +242,7 @@ public:
             if (not transient)
                 m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
         }
-        else if (cp and isdigit(*cp))
+        else if (auto cp = key.codepoint(); cp and isdigit(*cp))
         {
             long long new_val = (long long)m_params.count * 10 + *cp - '0';
             if (new_val > std::numeric_limits<int>::max())
@@ -284,7 +262,7 @@ public:
         }
         else if (key == '"')
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "register", KeymapMode::None,
                 [this](Key key, Context& context) {
                     auto cp = key.codepoint();
                     if (not cp or key == Key::Escape)
@@ -310,6 +288,15 @@ public:
             if (context().has_client())
                 context().client().info_hide();
 
+            // Hack to parse keys sent by terminals using the 8th bit to mark the
+            // meta key. In normal mode, give priority to a potential alt-key than
+            // the accentuated character.
+            if (key.key >= 127 and key.key < 256)
+            {
+                key.modifiers |= Key::Modifiers::Alt;
+                key.key &= 0x7f;
+            }
+
             do_restore_hooks = true;
             if (auto command = get_normal_command(key))
             {
@@ -490,7 +477,7 @@ public:
         }
         else if (key == alt('f'))
             to_next_word_begin<Word>(m_cursor_pos, m_line);
-        else if (key == alt('W'))
+        else if (key == alt('F'))
             to_next_word_begin<WORD>(m_cursor_pos, m_line);
         else if (key == alt('b'))
             to_prev_word_begin<Word>(m_cursor_pos, m_line);
@@ -514,7 +501,7 @@ public:
         else if (key == ctrl('w') or key == alt(Key::Backspace))
             erase_move(&to_prev_word_begin<Word>);
         else if (key == ctrl('W'))
-            erase_move(&to_prev_word_begin<Word>);
+            erase_move(&to_prev_word_begin<WORD>);
         else if (key == alt('d'))
             erase_move(&to_next_word_begin<Word>);
         else if (key == alt('D'))
@@ -789,8 +776,7 @@ public:
                                           completion);
             }
 
-            if (not context().history_disabled())
-                history_push(line);
+            history_push(line);
             context().print_status(DisplayLine{});
             if (context().has_client())
                 context().client().menu_hide();
@@ -806,8 +792,7 @@ public:
         }
         else if (key == Key::Escape or key == ctrl('c'))
         {
-            if (not context().history_disabled())
-                history_push(line);
+            history_push(line);
             context().print_status(DisplayLine{});
             if (context().has_client())
                 context().client().menu_hide();
@@ -821,7 +806,7 @@ public:
         }
         else if (key == ctrl('r'))
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "register", KeymapMode::None,
                 [this](Key key, Context&) {
                     auto cp = key.codepoint();
                     if (not cp or key == Key::Escape)
@@ -838,7 +823,7 @@ public:
         }
         else if (key == ctrl('v'))
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "raw-key", KeymapMode::None,
                 [this](Key key, Context&) {
                     if (auto cp = get_raw_codepoint(key))
                     {
@@ -1103,7 +1088,7 @@ private:
 
     void history_push(StringView entry)
     {
-        if (entry.empty() or
+        if (entry.empty() or context().history_disabled() or
             (m_flags & PromptFlags::DropHistoryEntriesWithBlankPrefix and
              is_horizontal_blank(entry[0_byte])))
             return;
@@ -1115,8 +1100,8 @@ private:
 class NextKey : public InputMode
 {
 public:
-    NextKey(InputHandler& input_handler, KeymapMode keymap_mode, KeyCallback callback)
-        : InputMode(input_handler), m_callback(std::move(callback)), m_keymap_mode(keymap_mode) {}
+    NextKey(InputHandler& input_handler, String name, KeymapMode keymap_mode, KeyCallback callback)
+        : InputMode(input_handler), m_name{std::move(name)}, m_callback(std::move(callback)), m_keymap_mode(keymap_mode) {}
 
     void on_key(Key key) override
     {
@@ -1134,9 +1119,10 @@ public:
 
     KeymapMode keymap_mode() const override { return m_keymap_mode; }
 
-    StringView name() const override { return "next-key"; }
+    StringView name() const override { return m_name; }
 
 private:
+    String m_name;
     KeyCallback m_callback;
     KeymapMode m_keymap_mode;
 };
@@ -1164,7 +1150,6 @@ public:
         last_insert().keys.clear();
         last_insert().disable_hooks = context().hooks_disabled();
         last_insert().count = count;
-        context().hooks().run_hook(Hook::InsertBegin, "", context());
         prepare(mode, count);
     }
 
@@ -1208,11 +1193,6 @@ public:
         }
         else if (key == Key::Escape or key == ctrl('c'))
         {
-            if (m_in_end)
-                throw runtime_error("asked to exit insert mode while running InsertEnd hook");
-            m_in_end = true;
-            context().hooks().run_hook(Hook::InsertEnd, "", context());
-
             m_completer.reset();
             pop_mode();
         }
@@ -1288,7 +1268,7 @@ public:
             insert(*cp);
         else if (key == ctrl('r'))
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "register", KeymapMode::None,
                 [this](Key key, Context&) {
                     auto cp = key.codepoint();
                     if (not cp or key == Key::Escape)
@@ -1317,7 +1297,7 @@ public:
         }
         else if (key == ctrl('x'))
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "explicit-completion", KeymapMode::None,
                 [this](Key key, Context&) {
                     if (key.key == 'f')
                         m_completer.explicit_file_complete();
@@ -1351,7 +1331,7 @@ public:
         }
         else if (key == ctrl('v'))
         {
-            on_next_key_with_autoinfo(context(), KeymapMode::None,
+            on_next_key_with_autoinfo(context(), "raw-insert", KeymapMode::None,
                 [this, transient](Key key, Context&) {
                     if (auto cp = get_raw_codepoint(key))
                     {
@@ -1507,7 +1487,6 @@ private:
     const bool      m_restore_cursor;
     bool            m_auto_complete;
     Timer           m_idle_timer;
-    bool            m_in_end = false;
     MouseHandler    m_mouse_handler;
     ScopedSetBool   m_disable_hooks;
 };
@@ -1531,7 +1510,7 @@ void InputHandler::push_mode(InputMode* new_mode)
     m_mode_stack.emplace_back(new_mode);
     new_mode->on_enabled();
 
-    context().hooks().run_hook(Hook::ModeChange, format("{}:{}", prev_name, new_mode->name()), context());
+    context().hooks().run_hook(Hook::ModeChange, format("push:{}:{}", prev_name, new_mode->name()), context());
 }
 
 void InputHandler::pop_mode(InputMode* mode)
@@ -1545,7 +1524,7 @@ void InputHandler::pop_mode(InputMode* mode)
     m_mode_stack.pop_back();
     current_mode().on_enabled();
 
-    context().hooks().run_hook(Hook::ModeChange, format("{}:{}", prev_name, current_mode().name()), context());
+    context().hooks().run_hook(Hook::ModeChange, format("pop:{}:{}", prev_name, current_mode().name()), context());
 }
 
 void InputHandler::reset_normal_mode()
@@ -1554,12 +1533,8 @@ void InputHandler::reset_normal_mode()
     if (m_mode_stack.size() == 1)
         return;
 
-    StringView prev_name = current_mode().name();
-    current_mode().on_disabled(false);
-    m_mode_stack.resize(1);
-    current_mode().on_enabled();
-
-    context().hooks().run_hook(Hook::ModeChange, format("{}:{}", prev_name, current_mode().name()), context());
+    while (m_mode_stack.size() > 1)
+        pop_mode(m_mode_stack.back().get());
 }
 
 void InputHandler::insert(InsertMode mode, int count)
@@ -1613,9 +1588,9 @@ void InputHandler::menu(Vector<DisplayLine> choices, MenuCallback callback)
     push_mode(new InputModes::Menu(*this, std::move(choices), std::move(callback)));
 }
 
-void InputHandler::on_next_key(KeymapMode keymap_mode, KeyCallback callback)
+void InputHandler::on_next_key(StringView mode_name, KeymapMode keymap_mode, KeyCallback callback)
 {
-    push_mode(new InputModes::NextKey(*this, keymap_mode, std::move(callback)));
+    push_mode(new InputModes::NextKey(*this, format("next-key[{}]", mode_name), keymap_mode, std::move(callback)));
 }
 
 InputHandler::ScopedForceNormal::ScopedForceNormal(InputHandler& handler, NormalParams params)
@@ -1747,7 +1722,7 @@ void hide_auto_info_ifn(const Context& context, bool hide)
         context.client().info_hide();
 }
 
-void scroll_window(Context& context, LineCount offset)
+void scroll_window(Context& context, LineCount offset, bool adapt_cursor)
 {
     Window& window = context.window();
     Buffer& buffer = context.buffer();
@@ -1761,6 +1736,10 @@ void scroll_window(Context& context, LineCount offset)
 
     const LineCount line_count = buffer.line_count();
     win_pos.line = clamp(win_pos.line + offset, 0_line, line_count-1);
+    window.set_position(win_pos);
+
+    if (not adapt_cursor)
+        return;
 
     SelectionList& selections = context.selections();
     const BufferCoord cursor = selections.main().cursor();
@@ -1776,7 +1755,6 @@ void scroll_window(Context& context, LineCount offset)
                    buffer[line].length()-1);
 
     selections = SelectionList{buffer, BufferCoord{line, col}};
-    window.set_position(win_pos);
 }
 
 }
