@@ -74,9 +74,11 @@ static constexpr KeyAndName keynamemap[] = {
     { "pagedown", Key::PageDown },
     { "home", Key::Home },
     { "end", Key::End },
+    { "ins", Key::Insert },
     { "del", Key::Delete },
     { "plus", '+' },
     { "minus", '-' },
+    { "semicolon", ';' },
 };
 
 KeyList parse_keys(StringView str)
@@ -145,32 +147,24 @@ KeyList parse_keys(StringView str)
 
 String key_to_str(Key key)
 {
-    if (auto mouse_event = (key.modifiers & Key::Modifiers::MouseEvent))
+    const auto coord = key.coord() + DisplayCoord{1,1};
+    switch (key.modifiers)
     {
-        const auto coord = key.coord() + DisplayCoord{1,1};
-        switch ((Key::Modifiers)mouse_event)
-        {
-            case Key::Modifiers::MousePos:
-                return format("<mouse:move:{}.{}>", coord.line, coord.column);
-            case Key::Modifiers::MousePressLeft:
-                return format("<mouse:press_left:{}.{}>", coord.line, coord.column);
-            case Key::Modifiers::MousePressRight:
-                return format("<mouse:press_right:{}.{}>", coord.line, coord.column);
-            case Key::Modifiers::MouseReleaseLeft:
-                return format("<mouse:release_left:{}.{}>", coord.line, coord.column);
-            case Key::Modifiers::MouseReleaseRight:
-                return format("<mouse:release_right:{}.{}>", coord.line, coord.column);
-            case Key::Modifiers::MouseWheelDown:
-                return "<mouse:wheel_down>";
-            case Key::Modifiers::MouseWheelUp:
-                return "<mouse:wheel_up>";
-            default: kak_assert(false);
-        }
-    }
-    else if (key.modifiers == Key::Modifiers::Resize)
-    {
-        auto size = key.coord() + DisplayCoord{1,1};
-        return format("<resize:{}.{}>", size.line, size.column);
+        case Key::Modifiers::MousePos:
+            return format("<mouse:move:{}.{}>", coord.line, coord.column);
+        case Key::Modifiers::MousePressLeft:
+            return format("<mouse:press_left:{}.{}>", coord.line, coord.column);
+        case Key::Modifiers::MousePressRight:
+            return format("<mouse:press_right:{}.{}>", coord.line, coord.column);
+        case Key::Modifiers::MouseReleaseLeft:
+            return format("<mouse:release_left:{}.{}>", coord.line, coord.column);
+        case Key::Modifiers::MouseReleaseRight:
+            return format("<mouse:release_right:{}.{}>", coord.line, coord.column);
+        case Key::Modifiers::Scroll:
+            return format("<scroll:{}>", static_cast<int>(key.key));
+        case Key::Modifiers::Resize:
+            return format("<resize:{}.{}>", coord.line, coord.column);
+        default: break;
     }
 
     bool named = false;
@@ -182,7 +176,7 @@ String key_to_str(Key key)
         named = true;
         res = it->name;
     }
-    else if (key.key >= Key::F1 and key.key < Key::F12)
+    else if (key.key >= Key::F1 and key.key <= Key::F12)
     {
         named = true;
         res = "F" + to_string((int)(key.key - Key::F1 + 1));
