@@ -10,6 +10,7 @@
 #include "vector.hh"
 #include "utils.hh"
 
+#include <cstdio>
 #include <cstring>
 
 namespace Kakoune
@@ -160,7 +161,7 @@ struct RegexParser
 private:
     struct InvalidPolicy
     {
-        Codepoint operator()(Codepoint cp) { throw regex_error{"Invalid utf8 in regex"}; }
+        Codepoint operator()(Codepoint cp) const { throw regex_error{"Invalid utf8 in regex"}; }
     };
 
     enum class Flags
@@ -348,6 +349,29 @@ private:
         }
     }
 
+    Codepoint read_hex(size_t count)
+    {
+        Codepoint res = 0;
+        for (int i = 0; i < count; ++i)
+        {
+            if (at_end())
+                parse_error("unterminated hex sequence");
+            Codepoint digit = *m_pos++;
+            Codepoint digit_value;
+            if ('0' <= digit and digit <= '9')
+                digit_value = digit - '0';
+            else if ('a' <= digit and digit <= 'f')
+                digit_value = 0xa + digit - 'a';
+            else if ('A' <= digit and digit <= 'F')
+                digit_value = 0xa + digit - 'A';
+            else
+                parse_error(format("invalid hex digit '{}'", digit));
+
+            res = res * 16 + digit_value;
+        }
+        return res;
+    }
+
     NodeIndex atom_escape()
     {
         const Codepoint cp = *m_pos++;
@@ -380,29 +404,6 @@ private:
                 return new_node(ParsedRegex::Literal, control.value);
         }
 
-        auto read_hex = [this](size_t count)
-        {
-            Codepoint res = 0;
-            for (int i = 0; i < count; ++i)
-            {
-                if (at_end())
-                    parse_error("unterminated hex sequence");
-                Codepoint digit = *m_pos++;
-                Codepoint digit_value;
-                if ('0' <= digit and digit <= '9')
-                    digit_value = digit - '0';
-                else if ('a' <= digit and digit <= 'f')
-                    digit_value = 0xa + digit - 'a';
-                else if ('A' <= digit and digit <= 'F')
-                    digit_value = 0xa + digit - 'A';
-                else
-                    parse_error(format("invalid hex digit '{}'", digit));
-
-                res = res * 16 + digit_value;
-            }
-            return res;
-        };
-
         if (cp == '0')
             return new_node(ParsedRegex::Literal, '\0');
         else if (cp == 'c')
@@ -417,14 +418,14 @@ private:
         else if (cp == 'x')
             return new_node(ParsedRegex::Literal, read_hex(2));
         else if (cp == 'u')
-            return new_node(ParsedRegex::Literal, read_hex(4));
+            return new_node(ParsedRegex::Literal, read_hex(6));
 
         if (contains("^$\\.*+?()[]{}|", cp)) // SyntaxCharacter
             return new_node(ParsedRegex::Literal, cp);
         parse_error(format("unknown atom escape '{}'", cp));
     }
 
-    void normalize_ranges(Vector<CharacterClass::Range, MemoryDomain::Regex>& ranges)
+    static void normalize_ranges(Vector<CharacterClass::Range, MemoryDomain::Regex>& ranges)
     {
         if (ranges.empty())
             return;
@@ -469,6 +470,20 @@ private:
             if (at_end())
                 break;
 
+            auto read_escaped_char = [this]() {
+                Codepoint cp = *m_pos++;
+                auto it = find_if(control_escapes, [cp](auto&& t) { return t.name == cp; });
+                if (it != std::end(control_escapes))
+                    return it->value;
+                if (cp == 'x')
+                    return read_hex(2);
+                if (cp == 'u')
+                    return read_hex(6);
+                if (not contains("^$\\.*+?()[]{}|-", cp)) // SyntaxCharacter and -
+                    parse_error(format("unknown character class escape '{}'", cp));
+                return cp;
+            };
+
             if (cp == '\\')
             {
                 auto it = find_if(character_class_escapes,
@@ -480,14 +495,7 @@ private:
                     continue;
                 }
                 else // its an escaped character
-                {
-                    cp = *m_pos++;
-                    auto it = find_if(control_escapes, [cp](auto&& t) { return t.name == cp; });
-                    if (it != std::end(control_escapes))
-                        cp = it->value;
-                    else if (not contains("^$\\.*+?()[]{}|-", cp)) // SyntaxCharacter and -
-                        parse_error(format("unknown character class escape '{}'", cp));
-                }
+                    cp = read_escaped_char();
             }
 
             CharacterClass::Range range = { cp, cp };
@@ -497,7 +505,10 @@ private:
                     break;
                 if (*m_pos != ']')
                 {
-                    range.max = *m_pos++;
+                    cp = *m_pos++;
+                    if (cp == '\\')
+                        cp = read_escaped_char();
+                    range.max = cp;
                     if (range.min > range.max)
                         parse_error("invalid range specified");
                 }
@@ -684,6 +695,7 @@ struct RegexCompiler
         {
             m_program.forward_start_desc = compute_start_desc<RegexMode::Forward>();
             compile_node<RegexMode::Forward>(0);
+            optimize(0, m_program.instructions.size());
             push_inst(CompiledRegex::Match);
         }
 
@@ -692,6 +704,7 @@ struct RegexCompiler
             m_program.first_backward_inst = m_program.instructions.size();
             m_program.backward_start_desc = compute_start_desc<RegexMode::Backward>();
             compile_node<RegexMode::Backward>(0);
+            optimize(m_program.first_backward_inst, m_program.instructions.size());
             push_inst(CompiledRegex::Match);
         }
         else
@@ -1019,6 +1032,43 @@ private:
         return std::make_unique<CompiledRegex::StartDesc>(start_desc);
     }
 
+    void optimize(size_t begin, size_t end)
+    {
+        if (not (m_flags & RegexCompileFlags::Optimize))
+            return;
+
+        auto is_jump = [](CompiledRegex::Op op) { return op >= CompiledRegex::Op::Jump and op <= CompiledRegex::Op::Split_PrioritizeChild; };
+        for (auto i = begin; i < end; ++i)
+        {
+            auto& inst = m_program.instructions[i];
+            if (is_jump(inst.op))
+                m_program.instructions[inst.param].last_step = 0xffff; // tag as jump target
+        }
+
+        for (auto block_begin = begin; block_begin < end; )
+        {
+            auto block_end = block_begin + 1;
+            while (block_end < end and
+                   not is_jump(m_program.instructions[block_end].op) and
+                   m_program.instructions[block_end].last_step != 0xffff)
+                ++block_end;
+            peephole_optimize(block_begin, block_end);
+            block_begin = block_end;
+        }
+    }
+
+    void peephole_optimize(size_t begin, size_t end)
+    {
+        // Move saves after all assertions on the same character
+        auto is_assertion = [](CompiledRegex::Op op) { return op >= CompiledRegex::LineStart; };
+        for (auto i = begin, j = begin + 1; j < end; ++i, ++j)
+        {
+            if (m_program.instructions[i].op == CompiledRegex::Save and
+                is_assertion(m_program.instructions[j].op))
+                std::swap(m_program.instructions[i], m_program.instructions[j]);
+        }
+    }
+
     const ParsedRegex::Node& get_node(ParsedRegex::NodeIndex index) const
     {
         return m_parsed_regex.nodes[index];
@@ -1129,7 +1179,7 @@ String dump_regex(const CompiledRegex& program)
                 res += "match\n";
         }
     }
-    auto dump_start_desc = [&](CompiledRegex::StartDesc& desc, StringView name) {
+    auto dump_start_desc = [&](const CompiledRegex::StartDesc& desc, StringView name) {
         res += name + " start desc: [";
         for (size_t c = 0; c < CompiledRegex::StartDesc::count; ++c)
         {
@@ -1521,13 +1571,32 @@ auto test_regex = UnitTest{[]{
         kak_assert(vm.exec("bCa"));
     }
 
+    {
+        TestVM<> vm{R"([\t-\r]+)"};
+        kak_assert(vm.exec("\t\n\v\f\r"));
+    }
+
+    {
+        TestVM<> vm{R"([^\x00-\x7F]+)"};
+        kak_assert(not vm.exec("ascii"));
+        kak_assert(vm.exec("←↑→↓"));
+        kak_assert(vm.exec("😄😊😉"));
+    }
+
+    {
+        TestVM<> vm{R"([^\u000000-\u00ffff]+)"};
+        kak_assert(not vm.exec("ascii"));
+        kak_assert(not vm.exec("←↑→↓"));
+        kak_assert(vm.exec("😄😊😉"));
+    }
+
     {
         TestVM<RegexMode::Forward | RegexMode::Search> vm{R"(д)"};
         kak_assert(vm.exec("д", RegexExecFlags::None));
     }
 
     {
-        TestVM<> vm{R"(\0\x0A\u260e\u260F)"};
+        TestVM<> vm{R"(\0\x0A\u00260e\u00260F)"};
         const char str[] = "\0\n☎☏"; // work around the null byte in the literal
         kak_assert(vm.exec({str, str + sizeof(str)-1}));
     }
