@@ -7,23 +7,24 @@
 namespace Kakoune
 {
 
+static const HashMap<String, Codepoint> reg_names = {
+    { "slash", '/' },
+    { "dquote", '"' },
+    { "pipe", '|' },
+    { "caret", '^' },
+    { "arobase", '@' },
+    { "percent", '%' },
+    { "dot", '.' },
+    { "hash", '#' },
+    { "underscore", '_' },
+    { "colon", ':' }
+};
+
 Register& RegisterManager::operator[](StringView reg) const
 {
     if (reg.length() == 1)
         return (*this)[reg[0_byte]];
 
-    static const HashMap<String, Codepoint> reg_names = {
-        { "slash", '/' },
-        { "dquote", '"' },
-        { "pipe", '|' },
-        { "caret", '^' },
-        { "arobase", '@' },
-        { "percent", '%' },
-        { "dot", '.' },
-        { "hash", '#' },
-        { "underscore", '_' },
-        { "colon", ':' }
-    };
     auto it = reg_names.find(reg);
     if (it == reg_names.end())
         throw runtime_error(format("no such register: '{}'", reg));
@@ -47,4 +48,9 @@ void RegisterManager::add_register(Codepoint c, std::unique_ptr<Register> reg)
     reg_ptr = std::move(reg);
 }
 
+CandidateList RegisterManager::complete_register_name(StringView prefix, ByteCount cursor_pos) const
+{
+    return complete(prefix, cursor_pos, reg_names | transform([](auto& i) { return i.key; }) | gather<Vector<String>>());
+}
+
 }
