@@ -135,7 +135,7 @@ const DisplayBuffer& Window::update_display_buffer(const Context& context)
         }
     }
 
-    DisplayBuffer::LineList& lines = m_display_buffer.lines();
+    DisplayLineList& lines = m_display_buffer.lines();
     m_display_buffer.set_timestamp(buffer().timestamp());
     lines.clear();
 
@@ -193,8 +193,17 @@ void Window::set_dimensions(DisplayCoord dimensions)
     if (m_dimensions != dimensions)
     {
         m_dimensions = dimensions;
-        run_hook_in_own_context(Hook::WinResize, format("{}.{}", dimensions.line,
-                                                    dimensions.column));
+        m_resize_hook_pending = true;
+    }
+}
+
+void Window::run_resize_hook_ifn()
+{
+    if (m_resize_hook_pending)
+    {
+        m_resize_hook_pending = false;
+        run_hook_in_own_context(Hook::WinResize,
+                                format("{}.{}", m_dimensions.line, m_dimensions.column));
     }
 }
 
@@ -280,7 +289,7 @@ ColumnCount find_display_column(const DisplayLine& line, const Buffer& buffer,
 BufferCoord find_buffer_coord(const DisplayLine& line, const Buffer& buffer,
                               ColumnCount column)
 {
-    auto& range = line.range();
+    const auto& range = line.range();
     for (auto& atom : line)
     {
         ColumnCount len = atom.length();
